/*
Copyright (c) 2002,2003, Christian Nowak <chnowak@web.de>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are 
permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of 
  conditions and the following disclaimer. 
- Redistributions in binary form must reproduce the above copyright notice, this list 
  of conditions and the following disclaimer in the documentation and/or other 
  materials provided with the distribution. 
- The names of the contributors may not be used to endorse or promote products derived 
  from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gpmm.h"
#include "gpos_def.h"
#include "gpmem.h"
#include "gpmodplay.h"

extern unsigned int bpmtab[224];

/* Global variables */
volatile BOOL DoMODPlay = TRUE;
static volatile unsigned int frame;
static volatile unsigned int * soundPos;
static volatile int idx_buf;
static unsigned int shiftVal;
static void * buffer;
void * gpmodplaybuffer;
static BOOL soundBufPlaying = FALSE;
static MODSOUNDBUF soundBuf;

/* This routine gets called by the timer interrupt and
 * polls the current playing position within the buffer.
 */
static void soundtimer ( void )
  {
    unsigned int t;

    if (!DoMODPlay)
      return;

    t = (((unsigned int)(*soundPos) - (unsigned int)buffer)>>shiftVal) >= soundBuf.samples ? 1 : 0;
    if (t!=frame)
      {
        unsigned int offs = ((frame==1) ? (soundBuf.samples<<shiftVal) : 0);
        soundBuf.callback(soundBuf.userdata, (u8*)((unsigned int)buffer+offs), soundBuf.samples<<shiftVal);
        frame = t;
      }
  }

static int SoundBufStart ( MODSOUNDBUF * sb, BOOL manual_polling )
  {
    frame = 0;

    if (soundBufPlaying)
      return -1;

    /* Copy the structure */
    gm_memcpy ( &soundBuf, sb, sizeof(MODSOUNDBUF) );

    /* Calculate size of a single sample in bytes
     * and a corresponding shift value
     */
    shiftVal = 0;
    switch (soundBuf.freq)
      {
        case PCM_S11:
        case PCM_S22:
        case PCM_S44:
          shiftVal++;
          break;
      }
    if (soundBuf.format == PCM_16BIT)
      shiftVal++;
    soundBuf.samplesize = 1<<shiftVal;

    /* Allocate memory for the playing buffer */
    buffer = gm_malloc ( soundBuf.samplesize*soundBuf.samples*2 );
    gpmodplaybuffer = buffer;
    gm_memset ( buffer, 0x80, soundBuf.samplesize*soundBuf.samples*2 );

    /* Set timer interrupt #0 */
    if (!manual_polling)
      {
        GpTimerOptSet ( 3, soundBuf.pollfreq, 0, soundtimer );
        GpTimerSet ( 3 );
      }

    /* Start playing */
    GpPcmInit ( soundBuf.freq, soundBuf.format );
    GpPcmPlay ( (unsigned short*)buffer, soundBuf.samples*soundBuf.samplesize*2, 1 );
    GpPcmLock ( (unsigned short*)buffer, (int*)&idx_buf, (unsigned int*)&soundPos );

    soundBufPlaying = TRUE;

    return 0;
  }


static void SoundBufStop ( BOOL manual_polling )
  {
    if (!soundBufPlaying)
      return;

    GpPcmStop();
    GpPcmRemove ( (unsigned short*)buffer );
    if (!manual_polling)
      GpTimerKill ( 3 );

    free ( buffer );
    
    soundBufPlaying = FALSE;
  }

static void mixcallback ( void * userdata, u8 * stream, int len )
  {
    int i;
    MODPlay * mp = (MODPlay*)userdata;
    MOD * mmod = &mp->mod;

    gpmodplaybuffer = stream;
    mmod->mixingbuf = stream;
    mmod->mixingbuflen = len;

    if (mp->paused)
      {
        for (i=0;i<len>>1;i++)
          ((u16*)(stream))[i] = 32768;
      } else
      {
        MOD_Player ( mmod );
      }
  }

static int updateWaveFormat(MODPlay * modplay)
  {
    BOOL p = modplay->playing;
    if (p)
      SoundBufStop(modplay->manual_polling);
    
    if (modplay->stereo)
      {
        if (modplay->playfreq==44100)
          modplay->soundBuf.freq = PCM_S44;
        else
        if (modplay->playfreq==22050)
          modplay->soundBuf.freq = PCM_S22;
        else
        if (modplay->playfreq==11025)
          modplay->soundBuf.freq = PCM_S11;
        else
          return -1;
        modplay->mod.channels = 2;
      } else
      {
        if (modplay->playfreq==44100)
          modplay->soundBuf.freq = PCM_M44;
        else
        if (modplay->playfreq==22050)
          modplay->soundBuf.freq = PCM_M22;
        else
        if (modplay->playfreq==11025)
          modplay->soundBuf.freq = PCM_M11;
        else
          return -1;
        modplay->mod.channels = 1;
      }
    modplay->mod.freq = modplay->playfreq;
    modplay->mod.bits = 16;   /* 8 bits are not supported yet */

    if (p)
      {
        modplay->mod.samplescounter = 0;
        modplay->mod.samplespertick = bpmtab[modplay->mod.bpm-32];
        if (modplay->mod.freq==22050)
          modplay->mod.samplespertick >>= 1;
        else
        if (modplay->mod.freq==11025)
          modplay->mod.samplespertick >>= 2;
      }

    modplay->soundBuf.format = PCM_16BIT;
    modplay->soundBuf.userdata = modplay;
    modplay->soundBuf.callback = mixcallback;
    modplay->soundBuf.samples = modplay->playfreq/(modplay->manual_polling?25:50);
    modplay->soundBuf.pollfreq = (modplay->playfreq/modplay->soundBuf.samples)*8;

    GpPcmInit ( modplay->soundBuf.freq, modplay->soundBuf.format );

    if (p)
      SoundBufStart(&modplay->soundBuf, modplay->manual_polling);
    
    return 0;
  }

void MODPlay_PollSound ( MODPlay * modplay )
  {
    if (!modplay->playing || !modplay->manual_polling)
      return;

    soundtimer();
  }

void MODPlay_Init ( MODPlay * modplay )
  {
    gm_memset ( modplay, 0, sizeof(MODPlay) );
    modplay->playing = FALSE;
    MODPlay_SetPlayFreq ( modplay, 44100 );
    MODPlay_SetStereo ( modplay, TRUE );
    modplay->paused = FALSE;
    modplay->bits = TRUE;
    modplay->numSFXChannels = 0;
    modplay->manual_polling = FALSE;
  }

int MODPlay_SetPlayFreq ( MODPlay * modplay, int f )
  {
    if ( f==modplay->playfreq )
      return 0;

    if ( f==44100 || f==22050 || f==11025 )
      {
        modplay->playfreq = f;
        updateWaveFormat(modplay);
        return 0;
      }
    else
      return -1;
  }

void MODPlay_SetStereo ( MODPlay * modplay, BOOL s )
  {
    if (s==modplay->stereo)
      return;

    modplay->stereo = s;
    updateWaveFormat(modplay);
  }

int MODPlay_AllocSFXChannels ( MODPlay * modplay, int sfxchans )
  {
    if (modplay->mod.modraw==NULL)
      return -1;
    
    if ( MOD_AllocSFXChannels(&modplay->mod, sfxchans) == 0 )
      {
        modplay->numSFXChannels = sfxchans;
        return 0;
      } else
      {
        return -1;
      }
  }

int MODPlay_Load ( MODPlay * modplay, char * fname )
  {
    BOOL playing;
    MOD tempmod;
    
    playing = modplay->playing;

    if ( MOD_Load ( &tempmod, fname ) < 0 )
      return -1;

    MODPlay_Unload(modplay);
    gm_memcpy ( &modplay->mod, &tempmod, sizeof(MOD) );
    MODPlay_AllocSFXChannels(modplay, modplay->numSFXChannels);

    if (playing)
      MODPlay_Start ( modplay );
    
    return 0;
  }

void MODPlay_Unload ( MODPlay * modplay )
  {
    MODPlay_Stop(modplay);
    MOD_Free(&modplay->mod);
  }

int MODPlay_Start ( MODPlay * modplay )
  {
    if (modplay->playing)
      return -1;
    if (modplay->mod.modraw==NULL)
      return -1;

    updateWaveFormat(modplay);
    MOD_Start ( &modplay->mod );
    if (SoundBufStart ( &modplay->soundBuf, modplay->manual_polling )<0)
      return -1;
    modplay->playing = TRUE;

    return 0;
  }

int MODPlay_Stop ( MODPlay * modplay )
  {
    if (!modplay->playing)
      return -1;

    SoundBufStop(modplay->manual_polling);
    modplay->playing = FALSE;
    return 0;
  }

int MODPlay_SetMOD ( MODPlay * modplay, u8 * mem )
  {
    MODPlay_Unload(modplay);
    
    if ( MOD_SetMOD ( &modplay->mod, mem ) == 0 )
      {
        MODPlay_AllocSFXChannels ( modplay, modplay->numSFXChannels );
        return 0;
      } else
      {
        return -1;
      }
  }

int MODPlay_Pause ( MODPlay * modplay, BOOL p )
  {
    if (!modplay->playing)
      return -1;

    modplay->paused = p;

    return 0;
  }

int MODPlay_TriggerNote ( MODPlay * modplay, int channel, u8 instnum, u16 freq, u8 vol )
  {
    if (modplay->mod.modraw==NULL)
      return -1;

    return MOD_TriggerNote ( &modplay->mod, channel, instnum, freq, vol );
  }

